module Item where

import Data.List
import ReadUtil

type Item = String

type Adj = String
type Noun = String

data Part =
    Whole Noun | Broken Part [Part]
    deriving (Show)

instance Eq Part where
    Whole x == Whole y =
        x == y
    Broken x xs == Broken y ys =
        (x == y) && (listsEqual xs ys)
    _ == _ =
        False

listsEqual :: (Eq a) => [a] -> [a] -> Bool
listsEqual [] [] = True
listsEqual (x:xs) ys =
    if x `elem` ys
    then listsEqual xs (delete x ys)
    else False
listsEqual _ _ = False

data ItemData = ItemData {
    item_name :: Item,
    item_adj :: Maybe Adj,
    item_noun :: Noun,
    item_part :: Part
} deriving (Show)

readsItem :: Item -> ReadS ItemData
readsItem name s =
    [(ItemData { item_name = name,
                 item_adj = findAdj sentences,
                 item_noun = noun,
                 item_part = findPart noun sentences },
      s') |
     (sentences, s1) <- readsSentences s,
     let noun = findNoun sentences,
     (_, s') <- readsPrompt s1]

findNoun :: [[String]] -> Noun
findNoun (("The":rest):_) = findNoun' rest

findNoun' :: [String] -> Noun
findNoun' ws = unwords (findNoun'' ws)

findNoun'' :: [String] -> [String]
findNoun'' ("is":_) = []
findNoun'' (w:ws) = w:(findNoun'' ws)

findAdj :: [[String]] -> Maybe Adj
findAdj [] = Nothing
findAdj (s:ss) =
    case parseAdj s of
      Just adj -> Just adj
      Nothing -> findAdj ss

parseAdj :: [String] -> Maybe Adj
parseAdj ("Interestingly":"this":"one":"is":rest) =
    Just (unwords rest)
parseAdj _ = Nothing

findPart :: Noun -> [[String]] -> Part
findPart _ [] = error "Couldn't find part"
findPart noun (s:ss) =
    case parsePart noun s of
      Just part -> part
      Nothing -> findPart noun ss

parsePart :: Noun -> [String] -> Maybe Part
parsePart noun ["Also", "it", "is", "in", "pristine", "condition"] =
    Just (Whole noun)
parsePart _ ("Also":"it":"is":"broken:":"it":"is":rest) =
    parsePart' (fixParens rest)
parsePart _ _ = Nothing

fixParens :: [String] -> [String]
fixParens ws =
    concatMap fixParens' ws

fixParens' :: String -> [String]
fixParens' w =
    groupBy fn w
    where
      fn '(' _ = False
      fn ')' _ = False
      fn _ '(' = False
      fn _ ')' = False
      fn _ _ = True

parsePart' :: [String] -> Maybe Part
parsePart' s =
    case suckPart (["("] ++ s ++ [")"]) of
      [(part, [])] -> Just part
      _ -> Nothing

suckPart :: [String] -> [(Part, [String])]
suckPart s =
    [(part, s') |
     (part, s') <- suckSimplePart s]
    ++
    [(Broken part parts, s') |
     (part, s1) <- suckSimplePart s,
     ("missing", s2) <- suckWord s1,
     (parts, s') <- suckPartList s2]

suckSimplePart :: [String] -> [(Part, [String])]
suckSimplePart s =
    [(Whole noun, s') |
     (noun, s') <- suckNoun s]
    ++
    [(part, s') |
     ("(", s1) <- suckWord s,
     (part, s2) <- suckPart s1,
     (")", s') <- suckWord s2]

suckNoun :: [String] -> [(Noun, [String])]
suckNoun s =
    [(unwords nounWords, s') |
     ((), s1) <- suckA s,
     (nounWords, s') <- suckNounWords s1,
     not (null nounWords)]

suckA :: [String] -> [((), [String])]
suckA s =
    [((), s') |
     ("a", s') <- suckWord s]
    ++
    [((), s') |
     ("an", s') <- suckWord s]

suckNounWords :: [String] -> [([String], [String])]
suckNounWords s@[] = [([], s)]
suckNounWords s@("a":_) = [([], s)]
suckNounWords s@("an":_) = [([], s)]
suckNounWords s@("missing":_) = [([], s)]
suckNounWords s@("and":_) = [([], s)]
suckNounWords s@("(":_) = [([], s)]
suckNounWords s@(")":_) = [([], s)]
suckNounWords (w:ws) =
    case suckNounWords ws of
      [(rest, s')] -> [((w:rest), s')]

suckPartList :: [String] -> [([Part], [String])]
suckPartList s =
    [([part], s') |
     (part, s') <- suckSimplePart s]
    ++
    [(part:parts, s') |
     (part, s1) <- suckSimplePart s,
     ("and", s2) <- suckWord s1,
     (parts, s') <- suckPartList s2]

suckWord :: [String] -> [(String, [String])]
suckWord [] = []
suckWord (w:ws) = [(w, ws)]
