module Solve where

import Data.List
import Item

data Tree =
    Use Item | Combine Tree [Tree]
    deriving (Show, Eq)

solveCommands :: Part -> [ItemData] -> [String]
solveCommands part itemDatas =
    let trees = map fst $ makeTrees part itemDatas
        solutions = map (solveTree itemDatas) trees
        sorted = sortBy sortSnd solutions
    in if null solutions
       then error "No solutions found"
       else fst $ head sorted --solutions
    where
      sortSnd a b = compare (snd a) (snd b)

solveTree :: [ItemData] -> Tree -> ([String], Int)
solveTree itemDatas tree =
    solveTree' itemDatas tree []

solveTree' :: [ItemData] -> Tree -> [Item] -> ([String], Int)
solveTree' [] tree inven =
    ([], length inven)
solveTree' (itemData:itemDatas) tree inven =
    let item = item_name itemData
    in if needed item tree
       then let inven' = inven ++ [item]
                (simpCmds, tree', inven'') = simplify tree inven'
                (restCmds, restSize) = solveTree' itemDatas tree' inven''
            in (["take " ++ item] ++ simpCmds ++ restCmds,
                max (length inven') restSize)
       else let (restCmds, restSize) = solveTree' itemDatas tree inven
            in (("take " ++ item):("inc " ++ item):restCmds,
                max (length inven + 1) restSize)

simplify :: Tree -> [Item] -> ([String], Tree, [Item])
simplify tree inven =
    simplify' tree inven pairs
    where
      pairs =
          [(item1, item2) |
           item1 <- inven,
           item2 <- inven,
           item1 /= item2]

simplify' :: Tree -> [Item] -> [(Item, Item)] -> ([String], Tree, [Item])
simplify' tree inven [] =
    ([], tree, inven)
simplify' tree inven (pair@(item1, item2):pairs) =
    if canSimplifyPair item1 item2 tree
    then let simplifyPairCmds = ["combine " ++ item1 ++ " with " ++ item2]
             tree' = simplifyPair item1 item2 tree
             inven' = delete item2 inven
             (nextCmds, tree'', inven'') = simplify tree' inven'
         in (simplifyPairCmds ++ nextCmds, tree'', inven'')
    else simplify' tree inven pairs

--data Tree =
--    Use Item | Combine Tree [Tree]

canSimplifyPair :: Item -> Item -> Tree -> Bool
canSimplifyPair _ _ (Use _) = False
canSimplifyPair item1 item2 (Combine tree trees) =
    ((tree == (Use item1)) &&
     ((Use item2) `elem` trees)) ||
    (canSimplifyPair item1 item2 tree) ||
    (any (canSimplifyPair item1 item2) trees)

simplifyPair :: Item -> Item -> Tree -> Tree
simplifyPair _ _ t@(Use item) = t
simplifyPair item1 item2 (Combine tree trees) =
    if tree == (Use item1)
       then let trees' = delete (Use item2) trees
            in if null trees'
               then tree
               else Combine tree trees'
       else Combine (simplifyPair item1 item2 tree) (map (simplifyPair item1 item2) trees)

needed :: Item -> Tree -> Bool
needed item1 (Use item2) =
    item1 == item2
needed item (Combine tree trees) =
    (needed item tree) || (any (needed item) trees)

makeTrees :: Part -> [ItemData] -> [(Tree, [ItemData])]
makeTrees part itemDatas =
    let options = map item_name $ filter (findPartInItemData part) itemDatas
    in concatMap (makeTrees' part itemDatas) options

makeTrees' :: Part -> [ItemData] -> Item -> [(Tree, [ItemData])]
makeTrees' part itemDatas item =
    let itemDatas' = deleteItem item itemDatas
        itemData = getItemData item itemDatas
        itemPart = item_part itemData
    in makeTrees'' part itemDatas' (Use item) itemPart

{-
makeTrees'' :: Part -> [ItemData] -> Item -> Part -> [(Tree, [ItemData])]
makeTrees'' part itemDatas item itemPart =
    if part == itemPart
    then [(Use item, itemDatas)]
    else case itemPart of
           Whole _ -> error "parts don't match in makeTrees''"
           Broken itemPart' missingParts ->
               [(Combine tree missingTrees, itemDatas'') |
                (tree, itemDatas') <-
                    makeTrees'' part itemDatas item itemPart',
                (missingTrees, itemDatas'') <-
                    makeTrees3 itemDatas' missingParts]
-}

makeTrees'' :: Part -> [ItemData] -> Tree -> Part -> [(Tree, [ItemData])]
makeTrees'' part itemDatas tree itemPart =
    if partsApproxEq part itemPart
    then case (part, itemPart) of
           (Broken x xs, Broken y ys)
               | (length xs < length ys) ->
                   let missingParts = partDiff xs ys
                   in [(tree', itemDatas') |
                       (missingTrees, itemDatas') <-
                           makeTrees3 itemDatas missingParts,
                       let tree' = Combine tree missingTrees]
           _ -> [(tree, itemDatas)]
    else case itemPart of
           Whole _ -> error "parts don't match in makeTrees''"
           Broken itemPart' missingParts ->
               [(tree'', itemDatas'') |
                (missingTrees, itemDatas') <-
                    makeTrees3 itemDatas missingParts,
                let tree' = Combine tree missingTrees,
                (tree'', itemDatas'') <-
                    makeTrees'' part itemDatas' tree' itemPart']


makeTrees3 :: [ItemData] -> [Part] -> [([Tree], [ItemData])]
makeTrees3 itemDatas [] = [([], itemDatas)]
makeTrees3 itemDatas (missingPart:missingParts) =
    [((tree:trees), itemDatas'') |
     (tree, itemDatas') <- makeTrees missingPart itemDatas,
     (trees, itemDatas'') <- makeTrees3 itemDatas' missingParts]
    

getItemData :: Item -> [ItemData] -> ItemData
getItemData item [] = error "getItemData: item not found"
getItemData item (itemData:itemDatas) =
    if item == item_name itemData
    then itemData
    else getItemData item itemDatas

deleteItem :: Item -> [ItemData] -> [ItemData]
deleteItem item [] = error "deleteItem: item not found"
deleteItem item (itemData:itemDatas) =
    if item == item_name itemData
    then itemDatas
    else itemData:(deleteItem item itemDatas)

findPartInItemData :: Part -> ItemData -> Bool
findPartInItemData part itemData =
    findPartInItemPart part (item_part itemData)

findPartInItemPart :: Part -> Part -> Bool
findPartInItemPart part itemPart =
    if partsApproxEq part itemPart
    then True
    else case itemPart of
           Whole _ -> False
           Broken itemPart' _ -> findPartInItemPart part itemPart'

partsApproxEq :: Part -> Part -> Bool
partsApproxEq (Whole x) (Whole y) = x == y
partsApproxEq (Broken x xs) (Broken y ys) =
    x == y && partSubset xs ys
partsApproxEq _ _ = False

partSubset :: [Part] -> [Part] -> Bool
partSubset [] _ = True
partSubset (x:xs) ys =
    (x `elem` ys) && (partSubset xs (delete x ys))

partDiff :: [Part] -> [Part] -> [Part]
partDiff [] ys = ys
partDiff (x:xs) ys =
    if (x `elem` ys)
    then partDiff xs (delete x ys)
    else error "partDiff called for not a proper subset"
