/* -*- mode:c++; c-style:k&r; c-basic-offset:4; indent-tabs-mode: nil; -*- */
/* vi:set expandtab cindent shiftwidth=4 cinoptions=\:0l1(0t0g0:           */


extern "C" {
#include <getopt.h>
#include <unistd.h>
}

#include <iostream>
#include <fstream>
#include <sstream>
#include <list>
#include <string>

using namespace std;

#include "world.hpp"


pair<world, brains> parse(const string &filename) {
    string name;
    int width = 0, height = 0;
    int y = 0;
    ifstream f(filename.c_str());
    getline(f, name);
    if (f.fail()) {
        cerr << "PARSE ERROR: no world name" << endl;
        exit(2);
    } else {
        cout << "World: " << name << endl;
    }
    bool reading_world = false;
    list<string> ant_brains;
    world *w = 0;
    int num_ants = 0;
    while (!f.eof()) {
        string line;
        getline(f, line, '\n');
        if (line == "") continue;
        if (!reading_world) {
            if ((line[0] >= '0') && (line[0] <= '9')) {
                reading_world = true;
                istringstream iss(line);
                iss >> width;
                iss >> height;
                w = new world(width, height);
            } else {
                ant_brains.push_back(line);
            }
        } else {
            assert(w);
            assert(y < height);
            for (int x = 0; x < width; x++) {
                char c1 = line[x+x];
                char c2 = line[x+x+1];
                cell_t cell;
                switch (c2) {
                case '#': cell = C_WALL; break;
                case 'o': cell = C_HOLE; break;
                case '$': cell = C_FOOD; break;
                case '-': cell = C_FLOOR; break;
                case '^': cell = ANT(c1 - '0', N); w->num_ants++; break;
                case '>': cell = ANT(c1 - '0', E); w->num_ants++; break;
                case 'v': cell = ANT(c1 - '0', S); w->num_ants++; break;
                case '<': cell = ANT(c1 - '0', W); w->num_ants++; break;
                default:
                    cerr << "ERROR: bad chars: " << c1 << c2 << endl;
                    assert(0);
                }
                w->set_cell(x,y,cell);
            }
            y++;
        }
    }
    brains b(ant_brains.size());
    int n = 0;
    for (list<string>::const_iterator i = ant_brains.begin();
         i != ant_brains.end(); i++, n++) {
        string s = *i;
        if (s.length() != 7) {
            cerr << "PARSE ERROR: invalid ant brain: " << s << endl;
            exit(2);
        }
        b.set_brain(n, s);
    }
    f.close();
    // leaks *w
    return pair<world,brains>(*w,b);
}

static struct option long_options[] = {
    { "interactive", required_argument, NULL, 'i' },
    { "help", no_argument, NULL, 'h' },
    { NULL, 0, NULL, 0 }
};

const string progname("antsim");

void usage(int exitcode) {
    ostream *pout = &cout;
    if (exitcode) pout = &cerr;
    ostream &out = *pout;
    out << progname << " [OPTION..] <ant_world>" << endl
        << "options:" << endl
        << "  -i --ineractive    pause at every step" << endl
        << "  -h --help          show usage" << endl;
    exit(exitcode);
}

int main(int argc, char **argv) {
    bool interactive = false;
    int opt;
    while ((opt = getopt_long(argc, argv, "ih", long_options, NULL)) != -1) {
        switch (opt) {
        case 'i': interactive = true; break;
        case 'h': usage(0); break;
        default:
            cerr << progname << ": unrecognized option: "
                 << argv[optind - 1] << endl;
            usage(2);
        }
    }
    argc -= optind;
    argv += optind;
    if (argc != 1) {
        usage(1);
    }
    pair<world, brains> wb = parse(argv[0]);
    world w = wb.first;
    brains b = wb.second;
    cout << "[2J";
    if (interactive) {
        b.show(cout);
        w.show(cout);
        cout << "enter = start simulation; q = exit" << endl;
    }
    int t;
    char c;
    for (t = 1; t < 100 & !w.done && w.num_ants > 0; t++) {
        if (interactive) {
            c = getchar();
            switch (c) {
            case 'q': 
            case 'Q': exit(0); break;
            }
        } else {
            usleep(100000);
        }
        w = step(b,w);
        cout << "[2J";
        cout << "tick " << t << ";  " << w.num_ants << " ants" << endl;
        w.show(cout);
        if (interactive) {
            cout << "enter = continue; q = exit" << endl;
        }
    }
    if (w.done) {
        cout << "victory!" << endl;
    } else if (w.num_ants < 1) {
        cout << "all ants died :(" << endl;
    } else {
        cout << "stopping after " << t << " cycles" << endl;
    }
}
