/* -*- mode:c++; c-style:k&r; c-basic-offset:4; indent-tabs-mode: nil; -*- */
/* vi:set expandtab cindent shiftwidth=4 cinoptions=\:0l1(0t0g0:           */

// #define NDEBUG
#include <cassert>
#include <vector>
#include <iostream>

#include "world.hpp"

char pp_cell(const cell_t c) {
    switch (CELL_TYPE(c)) {
    case C_WALL: return '#';
    case C_HOLE: return 'o';
    case C_FOOD: return '$';
    case C_FLOOR: return '-';
    case C_ANT:
        switch (ANT_DIR(c)) {
        case N: return '^';
        case E: return '>';
        case S: return 'v';
        case W: return '<';
#ifndef NDEBUG
        default: assert(0);
#endif
        }
#ifndef NDEBUG
    default: assert(0);
#endif
    }
}

void pp_pat(ostream &out, cell_t c, cell_t n, cell_t e, cell_t s, cell_t w) {
    out << "   " << pp_cell(n) << endl
        << "  " << pp_cell(w) << pp_cell(c) << pp_cell(e) << endl
        << "   " << pp_cell(s) << endl;
}

world step(const brains &b, const world &curr) {
    //cerr << "STEP" << endl;
    world next(curr); // copies world
    //cerr << "CURR:" << endl;
    //curr.show(cerr);
    int num_ants = next.num_ants;
    for (int y = 0; y < curr.size_y; y++) {
        for (int x = 0; x < curr.size_x; x++) {
            //cerr << "NEXT (" << x << "," << y << "):" << endl;
            //next.show(cerr);
            cell_t c = curr.get_cell(x,y);
            // rule 0: immovable things stay put
            if (C_IS_IMMOVABLE(c)) {
                // next.set_cell(x,y,c);
                continue;
            }

            cell_t n = curr.get_cell(x,y-1);
            cell_t e = curr.get_cell(x+1,y);
            cell_t s = curr.get_cell(x,y+1);
            cell_t w = curr.get_cell(x-1,y);
            //cerr << "(" << x << "," << y << ")" << endl;
            //pp_pat(cerr, c, n, e, s, w);
            //cerr << "cur:" << endl;
            //curr.show(cerr);
            //cerr << "next:" << endl;
            //next.show(cerr);

            // rule 0: no interested ants in sight
            if (CELL_TYPE(c) != C_ANT
                && (CELL_TYPE(n) != C_ANT || ANT_DIR(n) != S)
                && (CELL_TYPE(e) != C_ANT || ANT_DIR(e) != W)
                && (CELL_TYPE(s) != C_ANT || ANT_DIR(s) != N)
                && (CELL_TYPE(w) != C_ANT || ANT_DIR(w) != E)) {
                // no interested ants; keep cell
                // next.set_cell(x,y,c);
                //cerr << "DEBUG: uninterested: " << pp_cell(c) << endl;
                continue;
            }
                
            // rule 1:
            //  -
            // *^*   -->   -
            //  *
            if (C_IS_ANT_N(c) && (C_IS_FLOOR(n) || C_IS_HOLE(n))) {
                next.set_cell(x,y,C_FLOOR);
                num_ants--;
                //cerr << "DEBUG: 1N: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_E(c) && (C_IS_FLOOR(e) || C_IS_HOLE(e))) {
                next.set_cell(x,y,C_FLOOR);
                num_ants--;
                //cerr << "DEBUG: 1E: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_S(c) && (C_IS_FLOOR(s) || C_IS_HOLE(s))) {
                next.set_cell(x,y,C_FLOOR);
                num_ants--;
                //cerr << "DEBUG: 1S: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_W(c) && (C_IS_FLOOR(w) || C_IS_HOLE(w))) {
                next.set_cell(x,y,C_FLOOR);
                num_ants--;
                //cerr << "DEBUG: 1W: " << pp_cell(c) << endl;
                continue;
            }

            // rule 2:
            //  #
            // *^*   -->   >
            //  *
            if (C_IS_ANT_N(c) && C_IS_WALL(n)) {
                next.set_cell(x,y,TURN_ANT(c,E));
                //cerr << "DEBUG: 2N: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_E(c) && C_IS_WALL(e)) {
                next.set_cell(x,y,TURN_ANT(c,S));
                //cerr << "DEBUG: 2E: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_S(c) && C_IS_WALL(s)) {
                next.set_cell(x,y,TURN_ANT(c,W));
                //cerr << "DEBUG: 2S: " << pp_cell(c) << endl;
                continue;
            }
            if (C_IS_ANT_W(c) && C_IS_WALL(w)) {
                next.set_cell(x,y,TURN_ANT(c,N));
                //cerr << "DEBUG: 2W: " << pp_cell(c) << endl;
                continue;
            }

            if (C_IS_FLOOR(c)) {
                // rule 3: (four ants)
                //  v
                // >-<   -->   -
                //  ^
                if (C_IS_ANT_S(n) && C_IS_ANT_W(e)
                    && C_IS_ANT_N(s) && C_IS_ANT_E(w)) {
                    // next.set_cell(x,y,C_FLOOR);
                    //cerr << "DEBUG: 3: " << pp_cell(c) << endl;
                    continue;
                }

                // rule 4: (three ants)
                //  *
                // >-<   -->   s.p3
                //  ^
                if (                    C_IS_ANT_W(e)
                    && C_IS_ANT_N(s) && C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(s, b.get_dir(ANT_CLAN(s), P3)));
                    num_ants++;
                    //cerr << "DEBUG: 4N: " << pp_cell(c) << endl;
                    continue;
                }
                if (   C_IS_ANT_S(n)
                    && C_IS_ANT_N(s) && C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(w, b.get_dir(ANT_CLAN(w), P3)));
                    num_ants++;
                    //cerr << "DEBUG: 4E: " << pp_cell(c) << endl;
                    continue;
                }
                if (   C_IS_ANT_S(n) && C_IS_ANT_W(e)
                                     && C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(n, b.get_dir(ANT_CLAN(n), P3)));
                    num_ants++;
                    //cerr << "DEBUG: 4S: " << pp_cell(c) << endl;
                    continue;
                }
                if (   C_IS_ANT_S(n) && C_IS_ANT_W(e)
                    && C_IS_ANT_N(s)                 ) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(e, b.get_dir(ANT_CLAN(e), P3)));
                    num_ants++;
                    //cerr << "DEBUG: 4W: " << pp_cell(c) << endl;
                    continue;
                }
                
                // rule 5: (two ants across)
                //  v
                // *-*   -->   -
                //  ^
                if ((C_IS_ANT_S(n) && C_IS_ANT_N(s))
                    || (C_IS_ANT_W(e) && C_IS_ANT_E(w))) {
                    // next.set_cell(x,y,C_FLOOR);
                    //cerr << "DEBUG: 5: " << pp_cell(c) << endl;
                    continue;
                }
                
                // rule 6: (two ants diagonally)
                //  *
                // >-*   -->   s.p2
                //  ^
                if (C_IS_ANT_N(s) && C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(s, b.get_dir(ANT_CLAN(s), P2)));
                    num_ants++;
                    //cerr << "DEBUG: 6N: " << pp_cell(c) << endl;
                    continue;
                }
                if (C_IS_ANT_S(n) && C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(w, b.get_dir(ANT_CLAN(w), P2)));
                    num_ants++;
                    //cerr << "DEBUG: 6E: " << pp_cell(w) << endl;
                    continue;
                }
                if (C_IS_ANT_S(n) && C_IS_ANT_W(e)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(n, b.get_dir(ANT_CLAN(n), P2)));
                    num_ants++;
                    //cerr << "DEBUG: 6S: " << pp_cell(c) << endl;
                    continue;
                }
                if (C_IS_ANT_W(e) && C_IS_ANT_N(s)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(e, b.get_dir(ANT_CLAN(e), P2)));
                    num_ants++;
                    //cerr << "DEBUG: 6W: " << pp_cell(c) << endl;
                    continue;
                }

                // rule 7: (one ant)
                //  *
                // *-*   -->   s.p1
                //  ^
                if (C_IS_ANT_N(s)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(s, b.get_dir(ANT_CLAN(s), P1)));
                    num_ants++;
                    //cerr << "DEBUG: 7N: " << pp_cell(c) << endl;
                    continue;
                }
                if (C_IS_ANT_E(w)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(w, b.get_dir(ANT_CLAN(w), P1)));
                    num_ants++;
                    //cerr << "DEBUG: 7E: " << pp_cell(c) << endl;
                    continue;
                }
                if (C_IS_ANT_S(n)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(n, b.get_dir(ANT_CLAN(n), P1)));
                    num_ants++;
                    //cerr << "DEBUG: 7S: " << pp_cell(c) << endl;
                    continue;
                }
                if (C_IS_ANT_W(e)) {
                    next.set_cell(x,y,
                                  TURN_ANT_BY(e, b.get_dir(ANT_CLAN(e), P1)));
                    num_ants++;
                    //cerr << "DEBUG: 7W: " << pp_cell(c) << endl;
                    continue;
                }
            } // if c is floor

            // otherwise c is an ant or food

            // rule 8:
            //  ^
            // *^*   -->   c.p4
            //  *
            if ((C_IS_ANT_N(c) && C_IS_ANT_N(n))
                || (C_IS_ANT_E(c) && C_IS_ANT_E(e))
                || (C_IS_ANT_S(c) && C_IS_ANT_S(s))
                || (C_IS_ANT_W(c) && C_IS_ANT_W(w))) {
                
                cell_t new_c = TURN_ANT_BY(c, b.get_dir(ANT_CLAN(c), P4));
                if (new_c != c) next.set_cell(x,y,new_c);
                              
                //cerr << "DEBUG: 8: " << pp_cell(c) << endl;
                continue;
            }

            // rule 9:
            //  >
            // *^*   -->   c.p5
            //  *
            if ((C_IS_ANT_N(c) && C_IS_ANT_E(n))
                || (C_IS_ANT_E(c) && C_IS_ANT_S(e))
                || (C_IS_ANT_S(c) && C_IS_ANT_W(s))
                || (C_IS_ANT_W(c) && C_IS_ANT_N(w))) {
                cell_t new_c = TURN_ANT_BY(c, b.get_dir(ANT_CLAN(c), P5));
                if (new_c != c) next.set_cell(x,y,new_c);
                                  
                    //cerr << "DEBUG: 9: " << pp_cell(c) << endl;
                    //cerr << "DEBUG: 9: clan " << ANT_CLAN(c)
                    //                         << "; cmd: " << b.get_dir(ANT_CLAN(c), P5)
                    //                         << "; ant': " << pp_cell(TURN_ANT_BY(c, b.get_dir(ANT_CLAN(c), 5)))
                    //                         << endl;
                continue;
            }

            // rule 10:
            //  v
            // *^*   -->   c.p6
            //  *
            if ((C_IS_ANT_N(c) && C_IS_ANT_S(n))
                || (C_IS_ANT_E(c) && C_IS_ANT_W(e))
                || (C_IS_ANT_S(c) && C_IS_ANT_N(s))
                || (C_IS_ANT_W(c) && C_IS_ANT_E(w))) {
                cell_t new_c = TURN_ANT_BY(c, b.get_dir(ANT_CLAN(c), P6));
                if (new_c != c) next.set_cell(x,y,new_c);
                              
                //cerr << "DEBUG: 10: " << pp_cell(c) << endl;
                continue;
            }

            // rule 11:
            //  <
            // *^*   -->   c.p7
            //  *
            if ((C_IS_ANT_N(c) && C_IS_ANT_W(n))
                || (C_IS_ANT_E(c) && C_IS_ANT_N(e))
                || (C_IS_ANT_S(c) && C_IS_ANT_E(s))
                || (C_IS_ANT_W(c) && C_IS_ANT_S(w))) {
                cell_t new_c = TURN_ANT_BY(c, b.get_dir(ANT_CLAN(c), P7));
                if (new_c != c) next.set_cell(x,y,new_c);
                //cerr << "DEBUG: 10: " << pp_cell(c) << endl;
                continue;
            }

            // rule 12: (fixed)
            //  $
            // *^*
            //  *
            if ((C_IS_ANT_N(c) && C_IS_FOOD(n))
                || (C_IS_ANT_E(c) && C_IS_FOOD(e))
                || (C_IS_ANT_S(c) && C_IS_FOOD(s))
                || (C_IS_ANT_W(c) && C_IS_FOOD(w))) {
                // next.set_cell(x,y,c);
                next.done = true;
                continue;
            }
#ifndef NDEBUG
            // no pattern recognized
            cerr << "BAD STEP PATTERN: at " << x << "," << y << endl;
            pp_pat(cerr, c, n, e, s, w);
            assert(0);
#endif
        } // for x
    } // for y
    next.num_ants = num_ants;
    return next;
}
