/* -*- mode:c++; c-style:k&r; c-basic-offset:4; indent-tabs-mode: nil; -*- */
/* vi:set expandtab cindent shiftwidth=4 cinoptions=\:0l1(0t0g0:           */

#ifndef _WORLD_HPP_
#define _WORLD_HPP_

// #define NDEBUG
#include <cassert>
#include <vector>
#include <iostream>

using namespace std;

typedef enum {
    C_WALL = 0,
    C_HOLE = 1,
    C_FOOD = 2,
    C_FLOOR = 3,
    C_ANT = 4
} ct_t;

typedef enum {
    P1 = 0,
    P2 = 1,
    P3 = 2,
    P4 = 3,
    P5 = 4,
    P6 = 5,
    P7 = 6
} pat_t;

typedef enum { N = 0, E = 1, S = 2, W = 3 } dir_t;

#define ADD_DIR(d,diff) (((d) + (diff)) & 0x3)

typedef uint32_t cell_t;

#define CELL_TYPE(c)  ((c) & 0xff)
#define ANT_CLAN(c)   (((c) >> 16) & 0xff)
#define ANT_DIR(c)    (((c) >> 8) & 0xff)
#define ANT(c,d)      (((c) << 16) | ((d) << 8) | C_ANT)

#define C_IS_IMMOVABLE(c) ((c) < C_FLOOR)
#define C_IS_ANT(c) (CELL_TYPE(c) == C_ANT)
#define C_IS_WALL(c) ((c) == C_WALL)
#define C_IS_HOLE(c) ((c) == C_HOLE)
#define C_IS_FOOD(c) ((c) == C_FOOD)
#define C_IS_FLOOR(c) ((c) == C_FLOOR)
#define C_IS_ANT_N(c) (((c) & 0xffff) == 0x0004)
#define C_IS_ANT_E(c) (((c) & 0xffff) == 0x0104)
#define C_IS_ANT_S(c) (((c) & 0xffff) == 0x0204)
#define C_IS_ANT_W(c) (((c) & 0xffff) == 0x0304)

#define TURN_ANT(c, d) (((c) & 0xffff00ff) | ((d) << 8))
#define TURN_ANT_BY(c, d) (((c) & 0xffff00ff) | ((((c) & 0x0300) + (d << 8)) & 0x0300))

typedef uint32_t brain_t;

#define BRAIN_DIR(b, n)  (((b) >> ((n) << 1)) & 0x3)
#define BRAIN(p1,p2,p3,p4,p5,p6,p7) ((p1) | ((p2) << 2) | ((p3) << 4) | ((p4) << 6) | ((p5) << 8) | ((p6) << 10) | ((p7) << 12))

extern char pp_cell(const cell_t c);

class world {
public:
    world(int new_size_x, int new_size_y) :
        size_x(new_size_x), size_y(new_size_y), done(false), cells(0),
        num_ants(0)
    {
        void *p = malloc(new_size_x * new_size_y * sizeof(cell_t));
        assert(p);
        cells = static_cast<cell_t *>(p);
        for(int i = 0; i < new_size_x * new_size_y; i++) {
            *(cells + i) = C_FLOOR;
        }
        cerr << "init size_x " << size_x << " size_y " << size_y
             << " cells " << endl;
    }

    world(const world &w)
        : size_x(w.size_x), size_y(w.size_y), done(w.done), cells(0),
          score(w.score),num_ants(w.num_ants) {
        void *p = malloc(size_x * size_y * sizeof(cell_t));
        assert(p);
        assert(w.cells);
        cells = static_cast<cell_t *>(p);
        memcpy(cells, w.cells, size_x * size_y * sizeof(cell_t));
        cerr << "copy size_x " << size_x << " size_y " << size_y
             << " cells " << cells << endl;
    }

    ~world() {
        free(cells);
    }

    world &operator=(const world &w) {
        size_x = w.size_x;
        size_y = w.size_y;
        done = w.done;
        score = w.score;
        num_ants = w.num_ants;
        cells = 0;
        void *p = malloc(size_x * size_y * sizeof(cell_t));
        assert(p);
        assert(w.cells);
        cells = static_cast<cell_t *>(p);
        memcpy(cells, w.cells, size_x * size_y * sizeof(cell_t));
        cerr << "copy size_x " << size_x << " size_y " << size_y
             << " cells " << cells << endl;
    }

    const cell_t get_cell(int x, int y) const {
        assert(cells);
        if ((x < 0) || (y < 0) || (x >= size_x) || (y >= size_y))
            return C_WALL;
        return *(cells + (size_x * y) + x);
    }

    void set_cell(int x, int y, cell_t c) {
        assert(cells);
        assert(x >= 0); assert(y >= 0);
        assert(x < size_x); assert(y < size_y);
        *(cells + (size_x * y) + x) = c;
    }

    void show(ostream &out) const {
        assert(cells);
        out << size_x << " " << size_y << endl;
        for (int y = 0; y < size_y; y++) {
            for (int x = 0; x < size_x; x++) {
                cell_t c = get_cell(x, y);
                if (CELL_TYPE(c) == C_ANT) {
                    out << ANT_CLAN(c);
                } else {
                    out << ' ';
                }
                out << pp_cell(c);
            } // for x
            out << endl;
        } // for y
    } // show()

    int size_x;
    int size_y;
    bool done;
    cell_t *cells;
    uint64_t score;
    int num_ants;
};

class brains {
public:
    brains(int new_num_clans)
        : num_clans(new_num_clans) {
      assert(new_num_clans > 0  && new_num_clans <= 10);
    }
    const dir_t get_dir(int clan, int pat) const {
        assert(clan >= 0);
        assert(clan < num_clans);
        assert(pat >= 0);
        assert(pat < 7);
        return static_cast<dir_t>(BRAIN_DIR(bs[clan], pat));
    }
    void set_brain(int clan, brain_t b) {
        assert(clan >= 0);
        assert(clan < num_clans);
        bs[clan] = b;
    }
    void set_brain(int clan, string s) {
        assert(clan >= 0);
        assert(clan < num_clans);
        assert(s.size() == 7);
        int d[7];
        for (int i = 0; i < 7; i++) {
            switch (s[i]) {
            case 'N': d[i] = N; break;
            case 'E': d[i] = E; break;
            case 'S': d[i] = S; break;
            case 'W': d[i] = W; break;
#ifndef NDEBUG
            default: assert(0);
#endif
            }
        }
        set_brain(clan, BRAIN(d[0], d[1], d[2], d[3], d[4], d[5], d[6]));
    }
    void show(ostream &out) const {
        for (int c = 0; c < num_clans; c++) {
            brain_t b = bs[c];
            for (int i = 0; i < 7; i++) {
                char chr;
                switch(BRAIN_DIR(b, i)) {
                case N: chr = 'N'; break;
                case E: chr = 'E'; break;
                case S: chr = 'S'; break;
                case W: chr = 'W'; break;
#ifndef NDEBUG
                default:
                    cerr << "BAD BRAIN: " << hex << b << "; DIR: "
                         << hex << BRAIN_DIR(b, i) << endl;
                    assert(0);
#endif
                }
                out << chr;
            } // for dirs
            out << endl;
        } // for clans
    }
    int num_clans;
    brain_t bs[10];
};

extern world step(const brains &b, const world &curr);

#endif // _WORLD_HPP_
