module Exec where

import System.Exit
import Data.Array.IO
import Data.Bits
import Data.Char
import IO
import Control.Monad
import qualified Data.Map as Map
import UniversalMachine
import Numeric
import Flags


exec :: Flags -> Machine -> IO Machine
exec flags machine =
    do let err_arr = error ("*** missing imem ***")
           imem = {-# SCC "exec/getmem" #-}
                  Map.findWithDefault err_arr 0 (arrays machine)
       word <- {-# SCC "exec/fetch" #-}
               readArray imem (pc machine)
       let instr = {-# SCC "exec/decode" #-}
                   decode word
           dump = {-# SCC "exec/time" #-}
                  time machine >= dump_start flags
       
       ({-# SCC "exec/dump" #-}           
        when dump (do hPutStrLn stderr ("==> time 0x" ++ showHex (time machine) " <==")
                      print_regs stderr machine
                      hPutStrLn stderr ("0x" ++ showHex (pc machine) ": " ++ show instr)))
       machine' <- {-# SCC "exec/instr" #-}
                   exec_instr flags (machine { time = (time machine + 1),
                                               pc = (pc machine) + 1 }) instr
       exec flags machine'

exec_instr :: Flags -> Machine -> Instr -> IO Machine
exec_instr flags machine (IImm { ra = a, val = v })  =
    {-# SCC "imm" #-}
    do writeArray (regs machine) a v
       return machine
exec_instr flags machine (IReg { op = OpMov, ra = a, rb = b, rc = c }) =
    {-# SCC "mov" #-}
    do val_c <- readArray (regs machine) c
       val <- readArray (regs machine) b
       when (val_c /= 0) (writeArray (regs machine) a val)
       return machine
exec_instr flags machine (IReg { op = OpStore, ra = a, rb = b, rc = c }) =
    {-# SCC "store" #-}
    do val_a <- readArray (regs machine) a
       val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       let err_arr = error ("*** store to bad array: " ++ show val_b ++ " ***")
           mem = Map.findWithDefault err_arr val_a (arrays machine)
       writeArray mem val_b val_c
       return machine
exec_instr flags machine (IReg { op = OpLoad, ra = a, rb = b, rc = c }) =
    {-# SCC "load" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       let err_arr = error ("*** load from bad array: " ++ show val_b ++ " ***")
           mem = Map.findWithDefault err_arr val_b (arrays machine)
       val <- readArray mem val_c
       writeArray (regs machine) a val
       return machine
exec_instr flags machine (IReg { op = OpAdd, ra = a, rb = b, rc = c }) =
    {-# SCC "add" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       writeArray (regs machine) a (val_b + val_c)
       return machine
exec_instr flags machine (IReg { op = OpMul, ra = a, rb = b, rc = c }) =
    {-# SCC "mul" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       writeArray (regs machine) a (val_b * val_c)
       return machine
exec_instr flags machine (IReg { op = OpDiv, ra = a, rb = b, rc = c }) =
    {-# SCC "div" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       when (val_c == 0) (error "*** division by 0 ***")
       writeArray (regs machine) a (val_b `div` val_c)
       return machine
exec_instr flags machine (IReg { op = OpNand, ra = a, rb = b, rc = c }) =
    {-# SCC "nand" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       writeArray (regs machine) a (complement (val_b .&. val_c))
       return machine
exec_instr flags machine (IReg { op = OpHalt, ra = a, rb = b, rc = c }) =
    {-# SCC "halt" #-}
    do when (verbose flags) (hPutStrLn stderr ("***HALT***"))
       exitWith ExitSuccess
exec_instr flags machine (IReg { op = OpAlloc, ra = a, rb = b, rc = c }) =
    {-# SCC "alloc" #-}
    do let id = next_array_id machine
       val_c <- readArray (regs machine) c
       mem <- newArray (0, val_c - 1) 0
       writeArray (regs machine) b id
       let machine' = machine { next_array_id = id + 1,
                                arrays = Map.insert id mem (arrays machine) }
       return machine'
exec_instr flags machine (IReg { op = OpFree, ra = a, rb = b, rc = c }) =
    {-# SCC "free" #-}
    do val_c <- readArray (regs machine) c
       return (machine { arrays = Map.delete val_c (arrays machine) })
exec_instr flags machine (IReg { op = OpOut, ra = a, rb = b, rc = c }) =
    {-# SCC "out" #-}
    do val_c <- readArray (regs machine) c
       when (val_c > 0xff)
                (error ("*** invalid output char: " ++ show val_c ++ " ***"))
       (outf, machine') <-
           case out_file machine of
             Just f -> return (f, machine)
             Nothing -> do let id = next_file_id machine
                               fn = "out" ++ show (next_file_id machine)
                           f <- openFile fn WriteMode
                           hSetBuffering f NoBuffering
                           return (f, machine { out_file = Just f,
                                                next_file_id = id + 1 })
       buffering <- hGetBuffering outf
       let char = chr $ fromEnum (val_c)
           printable = (char >= ' ' && char <= '~') || char == '\n'
       case (buffering, printable) of
         (NoBuffering, True) -> putChar char
         (NoBuffering, False) -> do putChar (chr 0x85)
                                    hSetBuffering outf (BlockBuffering Nothing)
         (_, _) -> return ()
       hPutChar outf char
       return machine'
exec_instr flags machine (IReg { op = OpIn, ra = a, rb = b, rc = c }) =
    {-# SCC "in" #-}
    do let eof_handler e | isEOFError e = return 0xffffffff
                         | otherwise = ioError e
       hFlush stdout
       hFlush stderr
       machine' <- case (out_file machine) of
                     Nothing -> return machine
                     Just f -> do hClose f
                                  return (machine { out_file = Nothing })
       char <- catch (getChar >>= return . toEnum . ord) eof_handler
       writeArray (regs machine') c char
       return machine'
exec_instr flags machine (IReg { op = OpJump, ra = a, rb = b, rc = c }) =
    {-# SCC "jump" #-}
    do val_b <- readArray (regs machine) b
       val_c <- readArray (regs machine) c
       let err_arr = error ("*** load from bad array: " ++ show val_b ++ " ***")
           mem = Map.findWithDefault err_arr val_b (arrays machine)
       mem' <- if val_b == 0
               then return mem
               else mapArray id mem
       let arrays' = Map.adjust (const mem') 0 (arrays machine)
           machine' = machine { pc = val_c, arrays = arrays' }
       return machine'
