module UniversalMachine where

import Data.Word
import Data.Bits
import Array
import Numeric
import IO
import Data.Array.IO
import qualified Data.Map as Map

type Platter = Word32
type PlatterArray = IOUArray Platter Platter

type Reg = Word32

data Machine = Machine {
    pc :: Platter,
    time :: Word64,
    regs :: IOUArray Reg Platter,
    arrays :: Map.Map Platter PlatterArray,
    next_array_id :: Platter,
    next_file_id :: Int,
    out_file :: Maybe Handle
}

print_regs handle m =
    do rs <- mapM (readArray (regs m)) [0..7]
       hPutStrLn handle $ concat [show_reg r ++ ":" ++ show_platter v ++ "  "
                                  | (r, v) <- zip [0..] rs]

data Op = OpMov | OpLoad | OpStore | OpAdd | OpMul | OpDiv | OpNand
        | OpHalt | OpAlloc | OpFree | OpOut | OpIn | OpJump
          deriving (Eq)

data Instr = IReg { op :: Op, ra :: Reg, rb :: Reg, rc :: Reg }
           | IImm { ra :: Reg, val :: Platter }
             deriving (Eq)

show_reg :: Reg -> String
show_reg r = 'r' : show r

show_platter :: Platter -> String
show_platter p = '0' : 'x' : showHex p ""

instance Show Instr where
    show (IImm { ra = r, val = v }) =
        show_reg r ++ " <- " ++ show_platter v
    show (IReg { op = OpMov, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- " ++ show_reg b ++ " if " ++ show_reg c
    show (IReg { op = OpLoad, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- mem#" ++ show_reg b ++ "[" ++ show_reg c ++ "]"
    show (IReg { op = OpStore, ra = a, rb = b, rc = c }) =
        "mem#" ++ show_reg a ++ "[" ++ show_reg b ++ "] <- " ++ show_reg c
    show (IReg { op = OpAdd, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- " ++ show_reg b ++ " + " ++ show_reg c
    show (IReg { op = OpMul, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- " ++ show_reg b ++ " * " ++ show_reg c
    show (IReg { op = OpDiv, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- " ++ show_reg b ++ " / " ++ show_reg c
    show (IReg { op = OpNand, ra = a, rb = b, rc = c }) =
        show_reg a ++ " <- " ++ show_reg b ++ " ~& " ++ show_reg c
    show (IReg { op = OpHalt, ra = a, rb = b, rc = c }) = "halt"
    show (IReg { op = OpAlloc, ra = a, rb = b, rc = c }) =
        show_reg b ++ " <- alloc " ++ show_reg c
    show (IReg { op = OpFree, ra = a, rb = b, rc = c }) =
        "free " ++ show_reg c
    show (IReg { op = OpOut, ra = a, rb = b, rc = c }) =
        "out " ++ show_reg c
    show (IReg { op = OpIn, ra = a, rb = b, rc = c }) =
        show_reg c ++ " <- in"
    show (IReg { op = OpJump, ra = a, rb = b, rc = c }) =
        "j " ++ show_reg b ++ "@" ++ show_reg c

new_machine :: IO Machine
new_machine = 
    do rs <- newArray (0, 7) 0
       return (Machine { time = 0, pc = 0, regs = rs, arrays = Map.empty,
                         next_array_id = 1, next_file_id = 0, out_file = Nothing })

class Bittable t where
    decode :: Platter -> t

instance Bittable Op where
    decode 0 = OpMov
    decode 1 = OpLoad
    decode 2 = OpStore
    decode 3 = OpAdd
    decode 4 = OpMul
    decode 5 = OpDiv
    decode 6 = OpNand
    decode 7 = OpHalt
    decode 8 = OpAlloc
    decode 9 = OpFree
    decode 10 = OpOut
    decode 11 = OpIn
    decode 12 = OpJump
    decode n = error ("Unknown operator: " ++ show n)

instance Bittable Instr where
    decode word | op_num < 13 =
                    IReg { op = decode op_num,
                           ra = (word `shiftR` 6) .&. 0x7,
                           rb = (word `shiftR` 3) .&. 0x7,
                           rc = word .&. 0x7 }
                | op_num == 13 =
                    IImm { ra = (word `shiftR` 25) .&. 0x7,
                           val = word .&. 0x01ffffff }
        where op_num = (word `shiftR` 28) .&. 0xf
