#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include "jit.h"
#include "file.h"
#include "profile.h"
#include "codegen.h"
#include "fastalloc.h"

typedef unsigned int *(*mo_alloc_t)(unsigned int sz);
typedef void (*mo_free_t)(unsigned int *p);
typedef unsigned int (*mo_size_t)(unsigned int *p);

typedef struct {
  mo_alloc_t alloc;
  mo_free_t free;
  mo_size_t size;
} memops;

void usage()
{
  fprintf(stderr, "Usage: jit [-d|--debug] [-O<0-3>] <program>\n");
  exit(1);
}

int main(int argc, char *argv[])
{
  setbuf(stdout, (char *)NULL);
  setbuf(stderr, (char *)NULL);

  fa_init();

  memops mo;
  mo.alloc = fa_alloc;
  mo.free = fa_free;
  mo.size = fa_size;

  int debug = 0;
  const char *progname = 0;
  const char *opt = 0;

  unsigned int i;
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-h") == 0 ||
        strcmp(argv[i], "--help") == 0) {
      usage();
    } else if (strcmp(argv[i], "-d") == 0 ||
               strcmp(argv[i], "--debug") == 0) {
      debug = 1;
    } else if (strcmp(argv[i], "-O0") == 0 ||
               strcmp(argv[i], "-O") == 0 ||
               strcmp(argv[i], "-O1") == 0 ||
               strcmp(argv[i], "-O2") == 0 ||
               strcmp(argv[i], "-O3") == 0) {
      if (opt) {
        usage();
      } else {
        opt = argv[i];
      }
    } else {
      if (progname) {
        usage();
      } else {
        progname = argv[i];
      }
    }
  }
  if (!progname) {
    usage();
  }

  unsigned int *imem = 0;
  unsigned int imemsz = 0;
  readfile(progname, &imem, &imemsz);

  unsigned int pc = 0;

  unsigned int regs[8];
  for (i = 0; i < 8; i++) {
    regs[i] = 0;
  }

  int forcecompile = 1;

  unsigned int *oldinstrs = 0;
  unsigned int oldnuminstrs = 0;

  unsigned int *counts = 0;
  unsigned int *targets = 0;

  typedef int (*exec_t)(unsigned int **pimem,
                        unsigned int *pimemsz,
                        unsigned int *ppc,
                        unsigned int *pregs,
                        memops *mo);

  exec_t exec = NULL;
  void *lib = NULL;

  while (1) {
    unsigned int *instrs = 0;
    unsigned int numinstrs = 0;

    if (debug) {
      printf("Profiling...\n");
    }
    double value = profile(debug,
                           &imem,
                           &imemsz,
                           &pc,
                           regs,
                           &instrs,
                           &numinstrs,
                           oldinstrs,
                           oldnuminstrs,
                           &counts,
                           &targets);

    if (forcecompile || value > MINVALUE) {
      forcecompile = 0;

      if (oldinstrs) {
        free(oldinstrs);
      }
      oldinstrs = instrs;
      oldnuminstrs = numinstrs;

      if (debug) {
        printf("Generating code...\n");
      }
      codegen(imem, imemsz, instrs, numinstrs, targets);

      if (debug) {
        printf("Compiling...\n");
      }
      pid_t pid = fork();
      if (pid) {
        int status = 0;
        waitpid(pid, &status, 0);
        if (!WIFEXITED(status) ||
            WEXITSTATUS(status) != 0) {
          fprintf(stderr, "Compilation failed\n");
          exit(1);
        }
      } else {
        if (opt) {
          execlp(/*"time",*/ "gcc", "gcc",
                 "-fPIC", "-shared", "-Wall", "-g", opt,
                 "-o","exec.so", "exec.c", (char *)NULL);
        } else {
          execlp(/*"time",*/ "gcc", "gcc",
                 "-fPIC", "-shared", "-Wall", "-g",
                 "-o","exec.so", "exec.c", (char *)NULL);
        }
        fprintf(stderr, "Failed to execute gcc\n");
        exit(1);
      }

      if (lib) {
        dlclose(lib);
        lib = 0;
      }

      lib = dlopen("./exec.so", RTLD_LAZY);
      if (!lib) {
        fprintf(stderr, "Failed to load shared library\n");
        exit(1);
      }

      dlerror();
      exec = (exec_t)dlsym(lib, "exec");
      if (dlerror()) {
        fprintf(stderr, "Failed to locate symbol in shared library\n");
        exit(1);
      }
    } else {
      if (debug) {
        printf("Decided not to recompile.\n");
      }
    }

    if (debug) {
      printf("Executing compiled code...\n");
    }
    forcecompile = exec(&imem, &imemsz, &pc, regs, &mo);

    if (forcecompile) {
      if (oldinstrs) {
        free(oldinstrs);
        oldinstrs = 0;
      }
      oldnuminstrs = 0;
      if (counts) {
        free(counts);
        counts = 0;
      }
      if (targets) {
        free(targets);
        targets = 0;
      }
    }
  }

  return 0;
}
